#!/bin/bash

# env var defined in dockerfile / docker-compose :
# - BENCHMARKS_GO_DIR
# - BENCHMARKS_REF
# - BENCHMARKS_REF_REV
# - BENCHMARKS_ITERATION
# - BENCHMARKS_PATCH_DIR
# - BENCHMARKS_REVISIONS
# - BENCHMARKS_PROFILE
# - BENCHMARKS_REGEX
# - BENCHMARKS_REPORT_DIR

#set -x # debug

NB_CPU=$(LC_ALL=C lscpu | grep -E "^CPU\(s\):" | tr -s ' ' | cut -d ' ' -f 2)
BENCHMARKS_REPORT_DIR=${BENCHMARKS_REPORT_DIR:-/reports}

# move from relative to absolute path
if command -v realpath; then
        BENCHMARKS_REPORT_DIR=$(realpath "${BENCHMARKS_REPORT_DIR}")
else
        BENCHMARKS_REPORT_DIR=$(readlink -f "${BENCHMARKS_REPORT_DIR}")
fi

function failed() {
        if [ -z "$1" ]; then
                echo 'Error: Premature end of process: FAILURE!'
        else
                echo "Error: Premature end of process: $*"
        fi

        exit 1
}

run_benchmark() {
        FILE_PATH=$1

        if [ "${BENCHMARKS_REGEX}" == "" ]; then
                BENCHMARKS_REGEX="^Benchmark"
        fi

        PROFILE_OPTIONS=""
        if [ "${BENCHMARKS_PROFILE}" == "true" ]; then
                MEMPROFILE="$(dirname ${FILE_PATH})/$(basename ${FILE_PATH} .txt)-mem.profile"
                CPUPROFILE="$(dirname ${FILE_PATH})/$(basename ${FILE_PATH} .txt)-cpu.profile"
                PROFILE_OPTIONS="-memprofile ${MEMPROFILE} -cpuprofile ${CPUPROFILE}"
        fi

        go test -parallel ${NB_CPU} -benchmem -count=${BENCHMARKS_ITERATION} \
                ${PROFILE_OPTIONS} \
                -run=^$ -bench ${BENCHMARKS_REGEX} \
                ${BENCHMARKS_GO_DIR} | tee ${FILE_PATH} || failed "Unable to run benchmark '${BENCHMARKS_REGEX}'."
}

compute_ref() {
        IS_REF=$1
        GIT_REV=$2
        GIT_PATCH=$3

        echo "====================================================================="
        echo "==== Will run benchmarks for revision '$GIT_REV' (is_ref: $IS_REF)"

        PG_FEAT_DIR=/usr/src/pg_featureserv_cloned

        rm -rf ${PG_FEAT_DIR}

        if [ "${GIT_REV}" == "latest" ]; then
                # will fetch pg_featureserv in a local dir
                cp -a /usr/src/pg_featureserv_app/ ${PG_FEAT_DIR}/ || failed "Unable to copy latest sources."
        else
                # will fetch pg_featureserv from github
                cd /usr/src/ >/dev/null || failed "Unable to chdir"

                # clone ref repository
                git clone "https://github.com/benoitdm-oslandia/pg_featureserv.git" $PG_FEAT_DIR || failed "Unable to clone pg_featureserv repo."
                cd $PG_FEAT_DIR
                git reset --hard ${GIT_REV} || failed "Unable to find '${GIT_REV}' revision."

                # apply patch
                if [ -f ${BENCHMARKS_PATCH_DIR}/${GIT_PATCH} ]; then
                        git apply ${BENCHMARKS_PATCH_DIR}/${GIT_PATCH} || failed "Unable to apply patch."
                fi

                # update mods
                go mod tidy >/dev/null || failed "Unable to 'go mod tidy'."
                go mod verify >/dev/null || failed "Unable to 'go mod verify'."
        fi

        # copy benchmark go file to cloned repo as all pg_featureserv sources are in internal directory and are not accessible from outside package
        cp -a /usr/src/pg_bench_app/internal/* ${PG_FEAT_DIR}/internal/ || failed "Unable to copy benchmark sources."

        # goto pg_featureserv directory
        cd $PG_FEAT_DIR/ >/dev/null || failed "Unable to chdir"

        # install bench tools
        go install golang.org/x/perf/cmd/benchstat@latest >/dev/null || failed "Unable to install benchstat."

        # run benchmarks
        export PGFS_CACHE_TYPE="Disabled"
        run_benchmark ${BENCHMARKS_REPORT_DIR}/benchmarks-${GIT_REV}-no-cache.txt

        if [ "${BENCHMARKS_PROFILE}" == "true" ]; then
                MEMPROFILE="${BENCHMARKS_REPORT_DIR}/test-GetCollectionItemsDb-${GIT_REV}-mem.profile"
                CPUPROFILE="${BENCHMARKS_REPORT_DIR}/test-GetCollectionItemsDb-${GIT_REV}-cpu.profile"
                go test -parallel ${NB_CPU} \
                        -memprofile $MEMPROFILE -cpuprofile $CPUPROFILE \
                        -run GetCollectionItemsDb github.com/CrunchyData/pg_featureserv/internal/profiling \
                        || failed "Unable to run profiling."
        fi

        if [ "${IS_REF}" != "true" ]; then
                export PGFS_CACHE_TYPE="Naive"
                run_benchmark ${BENCHMARKS_REPORT_DIR}/benchmarks-${GIT_REV}-yes-cache.txt

                export PGFS_CACHE_TYPE="Redis"
                run_benchmark ${BENCHMARKS_REPORT_DIR}/benchmarks-${GIT_REV}-redis.txt
        fi

        # revert go.mod go.sum
        git restore go.mod go.sum

        echo "==== Done running benchmarks for revision '$GIT_REV'"
        echo "====================================================================="
}

zcat /usr/src/pg_bench_app/benchmarks/data.sql.gz | psql $DATABASE_URL >/dev/null || failed "Unable to inject default DB data."

BUILD_DATE=$(date +%d%m%Y-%H%M%S)

if [ "${BENCHMARKS_REF}" == "true" ]; then
        compute_ref true ${BENCHMARKS_REF_REV} "github-crunchydata-${BENCHMARKS_REF_REV}.patch"
fi

for rev in ${BENCHMARKS_REVISIONS}; do
        compute_ref false $rev "github-crunchydata-${rev}.patch"

        # ex. benchstat-23122023-15:03:56.txt
        REPORT_FILE="benchstat-${BENCHMARKS_REF_REV}-${rev}-${BUILD_DATE}.txt"

        benchstat ${BENCHMARKS_REPORT_DIR}/benchmarks-${BENCHMARKS_REF_REV}-no-cache.txt \
                ${BENCHMARKS_REPORT_DIR}/benchmarks-${rev}-no-cache.txt \
                ${BENCHMARKS_REPORT_DIR}/benchmarks-${rev}-redis.txt \
                ${BENCHMARKS_REPORT_DIR}/benchmarks-${rev}-yes-cache.txt \
                | tee ${BENCHMARKS_REPORT_DIR}/${REPORT_FILE} \
                || failed "Unable to run benchstat."

done
